(function(mod) {
  if (typeof exports == "object" && typeof module == "object") // CommonJS
    mod(require("../../lib/codemirror"), require("../../mode/sql/sql"));
  else if (typeof define == "function" && define.amd) // AMD
    define(["../../lib/codemirror", "../../mode/sql/sql"], mod);
  else // Plain browser env
    mod(CodeMirror);
})(function (CodeMirror) {
  function forEach(arr, f) {
    for (var i = 0, e = arr.length; i < e; ++i) f(arr[i]);
  }

  function arrayContains(arr, item) {
    if (!Array.prototype.indexOf) {
      var i = arr.length;
      while (i--) {
        if (arr[i] === item) {
          return true;
        }
      }
      return false;
    }
    return arr.indexOf(item) != -1;
  }

  function scriptHint(editor, _keywords, getToken) {
    // Find the token at the cursor
    var cur = editor.getCursor(), token = getToken(editor, cur), tprop = token;
    // If it's not a 'word-style' token, ignore the token.

    if (!/^[\w$_]*$/.test(token.string)) {
        token = tprop = {start: cur.ch, end: cur.ch, string: "", state: token.state,
                         className: token.string == ":" ? "python-type" : null};
    }

    if (!context) var context = [];
    context.push(tprop);

    var completionList = getCompletions(token, context);
    completionList = completionList.sort();
    //prevent autocomplete for last word, instead show dropdown with one word
    if(completionList.length == 1) {
      completionList.push(" ");
    }

    return {list: completionList,
            from: CodeMirror.Pos(cur.line, token.start),
            to: CodeMirror.Pos(cur.line, token.end)};
  }

  function pythonHint(editor) {
    return scriptHint(editor, pythonKeywordsU, function (e, cur) {return e.getTokenAt(cur);});
  }
  CodeMirror.pythonHint = pythonHint; // deprecated
  CodeMirror.registerHelper("hint", "python", pythonHint);

  var pythonKeywords = "False None True and as assert async await break class continue def del elif else " +
      "except finally for from global if import in is lambda nonlocal not or pass raise return try while with yield";
  var pythonKeywordsL = pythonKeywords.split(" ");
  var pythonKeywordsU = pythonKeywords.toUpperCase().split(" ");

  var pythonBuiltins = "ArithmeticError AssertionError AttributeError BaseException BlockingIOError BrokenPipeError " +
      "BufferError BytesWarning ChildProcessError ConnectionAbortedError ConnectionError " +
      "ConnectionRefusedError ConnectionResetError DeprecationWarning EOFError Ellipsis " +
      "EnvironmentError Exception False FileExistsError FileNotFoundError FloatingPointError " +
      "FutureWarning GeneratorExit IOError ImportError ImportWarning IndentationError IndexError " +
      "InterruptedError IsADirectoryError KeyError KeyboardInterrupt LookupError MemoryError " +
      "ModuleNotFoundError NameError None NotADirectoryError NotImplemented NotImplementedError OSError " +
      "OverflowError PendingDeprecationWarning PermissionError ProcessLookupError RecursionError ReferenceError " +
      "ResourceWarning RuntimeError RuntimeWarning StopAsyncIteration StopIteration SyntaxError SyntaxWarning " +
      "SystemError SystemExit TabError TimeoutError True TypeError UnboundLocalError UnicodeDecodeError " +
      "UnicodeEncodeError UnicodeError UnicodeTranslateError UnicodeWarning UserWarning ValueError Warning " +
      "ZeroDivisionError __build_class__ __debug__ __doc__ __import__ __loader__ __name__ __package__ __spec__ " +
      "abs all any ascii bin bool breakpoint bytearray bytes callable chr classmethod compile complex copyright " +
      "credits delattr dict dir divmod enumerate eval exec exit filter float format frozenset " +
      "getattr globals hasattr hash help hex id input int isinstance issubclass iter len license " +
      "list locals map max memoryview min next object oct open ord pow print property quit range repr reversed round " +
      "set setattr slice sorted staticmethod str sum super tuple type vars zip";
  var pythonBuiltinsL = pythonBuiltins.split(" ").join("() ").split(" ");
  var pythonBuiltinsU = pythonBuiltins.toUpperCase().split(" ").join("() ").split(" ");

  function getCompletions(token, context) {
    var found = [], start = token.string;
    function maybeAdd(str) {
      if (str.indexOf(start) == 0 && !arrayContains(found, str)) found.push(str);
    }

    function gatherCompletions(_obj) {
        forEach(pythonBuiltinsL, maybeAdd);
        // forEach(pythonBuiltinsU, maybeAdd);
        forEach(pythonKeywordsL, maybeAdd);
        // forEach(pythonKeywordsU, maybeAdd);
    }

    if (context) {
      // If this is a property, see if it belongs to some object we can
      // find in the current environment.
      var obj = context.pop(), base;

      if (obj.type == "variable")
          base = obj.string;
      else if(obj.type == "variable-3")
          base = ":" + obj.string;

      while (base != null && context.length)
        base = base[context.pop().string];
      if (base != null) gatherCompletions(base);
    }
    return found;
  }
});